unit Unit1;

interface

uses
  Windows, SysUtils, Classes, Forms,
  Grids, math, Controls, StdCtrls;

const
  nbcnt = 4;   // quantity of adjacent cells - neighbours : 8 or 4 only !
               // 4 for ortho directions, 8 - add diagonal directions

type
  
  NTerType = (ttWater, ttSpot);

  tCellint = Byte;

  Tfield = class;

  tCell = class (TObject)
  private
    fx, fy : tCellint;
    ftertype: NTerType;
    fpot : smallint;
    fneighbours : array [0..nbcnt-1] of tCell;
    ffield : tfield;
    passed: boolean;   //--- used only by getsquare method
    curnb : byte;      //
    spothead : tCell;  //
    fsquare : integer; //---
    function GetNeighbour(index: byte): Tcell;
    function GetPot: byte;
    function GetPotcnt: byte;
    function getsquare: cardinal;
    procedure desttocoorddelta(aind: byte; out ay, ax: shortint);
  public
    color : cardinal;
    constructor create (ay, ax : tCellint; atertype : NTerType; afield : Tfield);
    function distance (aCell : tCell) : tCellint;
    function isspothead () : Boolean;
    property field : Tfield read ffield;
    property neighbour[index : byte]: Tcell read GetNeighbour ;
    property pot: byte read GetPot;
    property potcnt: byte read GetPotcnt;
    property square : cardinal read getsquare;
    property tertype : Ntertype read fTertype;
    property x : tCellint read fx;
    property y : tCellint read fy;
  end;

  fieldCellar = array of array of TCell;

  TField = class
  private
    fheight, fwidth : tCellint;
    ffield : fieldCellar;
    function getheight: tCellint;
    function getwidth: tCellint;
    function getCell(ay, ax : tCellint): tCell;
    procedure setCell(ay, ax : tCellint; aCell : tCell);
  public
    constructor create (aheight, awidth, fact : tCellint);
    destructor destroy () ; override;
    function isvalidcell(ay, ax : tCellint) : Boolean;
    procedure clearmap;
    procedure getspots(list : tstrings);
    procedure randomfill(fact: byte = 1);
    property Field  [row : tCellint ; column : tCellint] : tCell  read getCell write setCell ; default ;
    property height : tCellint read getheight write fheight;
    property width : tCellint read getwidth write fwidth;
  end;

  TForm1 = class(TForm)
    SG1: TStringGrid;
    LB1: TListBox;
    Bcalc: TButton;
    SBrand: TScrollBar;
    Brecreate: TButton;
    SBx: TScrollBar;
    SBy: TScrollBar;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure SG1DrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure BcalcClick(Sender: TObject);
    procedure BrecreateClick(Sender: TObject);
  private
    lock : boolean;
    field : Tfield;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

{ tCell }

constructor tCell.create(ay, ax: tCellint ; atertype : NTerType; afield : Tfield);
begin
  fx := ax;
  fy := ay;
  ftertype := atertype;
  fpot := -1;
  fsquare := -1;
  spothead := nil;
  ffield := afield;
end;

procedure Tcell.desttocoorddelta(aind: byte; out ay, ax: shortint);
begin
  ay := round(sin(2*pi*aind/nbcnt));
  ax := round(cos(2*pi*aind/nbcnt));
end;

function tCell.distance(aCell: tCell): tCellint;
begin
  Result :=  Max(Max(x, aCell.x)- Min(x, aCell.x), Max(y, aCell.y)- Min(y, aCell.y))
end;

function tCell.GetNeighbour(index: byte): Tcell;
var
  ay, ax : shortint;
begin
  if fneighbours[index] <> nil then
    begin
    Result := fneighbours[index];
    Exit;
    end;
  desttocoorddelta(index, ay, ax);
  if fField.isvalidcell(fy+ay, fx+ax) then
    result := Field[fy+ay, fx+ax]
  else result := self ;
  fneighbours[index] := Result;
end;

function tCell.GetPot: byte;
var
  i : Byte;
begin
  if fpot <> -1 then
    begin
      Result := fpot;
      Exit;
    end;
  result := 0;
  fpot := Result;
  if ftertype = ttwater then
    Exit;
  for i := 0 to nbcnt-1 do
    if (Neighbour[i] <> self) and (Neighbour[i].ftertype = ttspot) then
       result := Result or (1 shl i);
  fpot := Result;
end;

function tCell.GetPotcnt: byte;
var
  i : Byte;
begin
  result := 0;
  for i := 0 to nbcnt-1 do
    Inc(Result, (pot shr i) and 1);
end;

function tCell.getsquare: cardinal;
var
  list : tlist;
  acell : tCell;
  curitem : integer;
begin
  if spothead <> nil then
    begin
    if spothead <> Self  then
      begin
      result := spothead.fsquare;
      Exit;
      end
    else if fsquare <> -1 then
      begin
        Result := fsquare;
        Exit;
      end;
    end
  else if ftertype in [ttwater]  then
    begin
    result := 0;
    fsquare := 0;
    Exit;
    end;

  list := tlist.Create;
  passed := True;
  list.add(self);
  spothead := Self;
  color := RGB(random(128)+128, random(128)+128, random(255));
  curitem := 0;
  while curitem < list.Count do
    begin
    acell :=  tcell(list[curitem]);
    repeat
      if not acell.neighbour[aCell.curnb].passed
        and (aCell.neighbour[aCell.curnb].tertype in [ttspot]) then
        begin
        aCell.neighbour[aCell.curnb].passed:= True;
        acell.neighbour[aCell.curnb].spothead := Self;
        list.add(aCell.neighbour[acell.curnb]);
        end
      else Inc (acell.curnb);
    until aCell.curnb = nbcnt;
    inc(curitem);
    end;
  fsquare := list.Count;
  Result := fsquare;
  for curitem := 0 to list.Count-1 do
    begin
      tCell(list[curitem]).curnb := 0;
      tCell(list[curitem]).passed := false;
    end;
  list.Free;
end;

function tCell.isspothead: Boolean;
begin
  result := self = spothead;
end;

{ TField }

procedure Tfield.clearmap;
var
  i, j : word ;
begin
  for i := 0 to height - 1 do
    for j := 0 to width - 1 do
      freeandnil(ffield[i,j]);
end;

constructor TField.create(aheight, awidth, fact: tCellint);
begin
  setlength (ffield, aheight, awidth);
  height := aheight ;
  width := awidth ;
  randomfill(fact);
end;

destructor TField.destroy;
begin
  clearmap;
  ffield := nil;
  inherited;
end;

function TField.getCell(ay, ax: tCellint): tCell;
begin
  if isvalidcell(ay,ax) then
    result := ffield[ay,ax]
  else raise Erangeerror.Create(Format('out of the field, call to x%d y%d', [ax,ay]));
end;

function TField.getheight: tCellint;
begin
  Result := fheight
end;


procedure TField.getspots(list: tstrings);
var
  w, h, c, s : cardinal;
  res: array of Byte;
begin
  c := 0;
  s := width*height;
  SetLength(res, s);
  ZeroMemory(res, s);
  for w := 0 to width -1 do
    for h := 0 to height -1 do
      if Field[h,w].isspothead then
        begin
          Inc(c);
          Inc(res[Field[h,w].square]);
        end;
  list.Clear;
  list.Add(IntToStr(c));
  for w := 0 to s-1 do
    if res[w] > 0 then
      list.Add(Format('s%d %d',[w,res[w]]));
end;

function Tfield.getwidth: tCellint;
begin
  Result := fwidth
end;

function TField.isvalidcell(ay, ax : tCellint): Boolean;
begin
  result := (ay in [0..height-1]) and (ax in [0..width-1]);
end;

procedure TField.randomfill(fact: byte);
var
  w, h : cardinal;
  t, o : NTerType;
begin
  o := ntertype(0);
  for w := 0 to width -1 do
    for h := 0 to height -1 do
      begin
      t := ntertype(Random(fact+ord(High(ntertype))));
      if T > High(ntertype)
        then t := o; //ttwater previously was t := o
      o := t;
      field[h,w] := tCell.create(h, w, t, self);
      end;
end;

procedure TField.setCell(ay, ax: tCellint; aCell: tCell);
begin
  ffield[ay,ax].Free;
  ffield[ay,ax] := aCell;
end;

{TForm1}

procedure TForm1.FormCreate(Sender: TObject);
begin
  randomize;
  LB1.Columns          := 12;
//  SG1.ScrollBars       := ssNone;
  SG1.FixedCols        := 0;
  SG1.FixedRows        := 0;
  sg1.DefaultColWidth  := 35;
  sg1.DefaultRowHeight := 13;
  sg1.GridLineWidth    := 0;
  SBrand.Max           := 10;
  SBrand.Min           := 1;
  SBx.Max              := 30;
  SBx.Min              := 5;
  SBy.Max              := 71;
  SBy.Min              := 5;
  sbrand.ShowHint      := True;
  sbx.ShowHint         := True;
  sby.ShowHint         := True;
  sbrand.Hint          := 'Randomness';
  sbx.Hint             := 'Field width';
  sby.Hint             := 'Field height';
  Brecreate.Click;
end;

procedure TForm1.SG1DrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  s : string;
begin
  if lock then
    Exit;
with SG1.Canvas, field[ARow, ACol] do
  begin
  s:= Format({'p%d' + { } '%d' + {} 's' +{} '%d',[{pot, {}  potcnt,{} square]);
  case tertype of
    ttSpot : if spothead <> nil then
               Brush.Color := spothead.color
             else Brush.Color := $008080 ;  // clolive
    ttwater : Brush.Color := $FF0000;       //clblue
    end;
  FillRect(rect);
  TextOut(Rect.Left, Rect.Top, s);
  end;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  field.Free;
end;

procedure TForm1.BcalcClick(Sender: TObject);
begin
  field.getspots(LB1.Items);
end;

procedure TForm1.BrecreateClick(Sender: TObject);
begin
  lock := true;
  field.Free;
  field := Tfield.create(sby.Position, SBx.Position, sbrand.Position);
  SG1.ColCount := SBx.Position;
  SG1.rowCount := sby.Position;
  lock := false;
  SG1.Refresh;
  LB1.Clear;
end;

end.
